
# This file was generated by 'versioneer.py' (0.18) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.
from __future__ import absolute_import
import json

version_json = '''
{
 "date": "2021-11-25T15:11:44-0500",
 "dirty": false,
 "error": null,
 "full-revisionid": "3584072c4a38f8ea2d7e08fdb6db177aa82f0bbc",
 "version": "1.14.0"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)



"""
Tests for L{eliot.filter}.
"""
from __future__ import unicode_literals

import sys

from unittest import TestCase
from datetime import datetime
from io import BytesIO
import inspect

from .common import FakeSys
from .. import _bytesjson as json
from ..filter import EliotFilter, main, USAGE


class EliotFilterTests(TestCase):
    """
    Tests for L{EliotFilter}.
    """

    def test_expression(self):
        """
        For each item in the incoming sequence L{EliotFilter.run} calls
        L{EliotFilter._evaluate} with the item decoded from JSON, and writes the
        result to the output file as JSON.
        """
        f = BytesIO()
        efilter = EliotFilter("J", [b'"abcd"', b"[1, 2]"], f)
        efilter._evaluate = lambda expr: {"x": len(expr), "orig": expr}
        self.assertEqual(f.getvalue(), b"")
        efilter.run()
        self.assertEqual(
            f.getvalue(),
            json.dumps({"x": 4, "orig": "abcd"})
            + b"\n"
            + json.dumps({"x": 2, "orig": [1, 2]})
            + b"\n",
        )

    def evaluateExpression(self, expr, message):
        """
        Render a single message with the given expression using
        L{EliotFilter._evaluate}.
        """
        efilter = EliotFilter(expr, [], BytesIO())
        return efilter._evaluate(message)

    def test_J(self):
        """
        The expression has access to the decoded JSON message as C{J} in its
        locals.
        """
        result = self.evaluateExpression("J['a']", {"a": 123})
        self.assertEqual(result, 123)

    def test_otherLocals(self):
        """
        The expression has access to L{datetime} and L{timedelta} in its
        built-ins.
        """
        result = self.evaluateExpression(
            "isinstance(datetime.utcnow() - datetime.utcnow(), timedelta)", {}
        )
        self.assertEqual(result, True)

    def test_datetimeSerialization(self):
        """
        Any L{datetime} in results will be serialized using L{datetime.isoformat}.
        """
        dt = datetime(2012, 12, 31)
        f = BytesIO()
        EliotFilter("datetime(2012, 12, 31)", ["{}"], f).run()
        expected = json.dumps(dt.isoformat()) + b"\n"
        self.assertEqual(f.getvalue(), expected)

    def test_SKIP(self):
        """
        A result of C{SKIP} indicates nothing should be output.
        """
        f = BytesIO()
        EliotFilter("SKIP", [b'{"a": 123}'], f).run()
        self.assertEqual(f.getvalue(), b"")


class MainTests(TestCase):
    """
    Test cases for L{main}.
    """

    def test_default(self):
        """
        By default L{main} uses information from L{sys}.
        """
        self.assertEqual(inspect.getargspec(main).defaults, (sys,))

    def test_stdinOut(self):
        """
        L{main} reads from the C{stdin} attribute of the given C{sys} equivalent,
        and writes rendered expressions to the C{stdout} attribute.
        """
        sys = FakeSys(["eliotfilter", "J[0]"], b"[1, 2]\n[4, 5]\n")
        main(sys)
        self.assertEqual(sys.stdout.getvalue(), b"1\n4\n")

    def test_success(self):
        """
        A successful run returns C{0}.
        """
        sys = FakeSys(["eliotfilter", "J[0]"], b"[1, 2]\n[4, 5]\n")
        result = main(sys)
        self.assertEqual(result, 0)

    def test_noArguments(self):
        """
        If given no arguments, usage documentation is printed to stderr and C{1}
        is returned.
        """
        sys = FakeSys(["eliotfilter"], b"")
        result = main(sys)
        self.assertEqual(sys.stderr.getvalue(), USAGE)
        self.assertEqual(result, 1)



"""
Tests for L{eliot._traceback}.
"""

from __future__ import unicode_literals

from unittest import TestCase, SkipTest
import traceback
import sys

try:
    from twisted.python.failure import Failure
except ImportError:
    Failure = None

from .._traceback import write_traceback, writeFailure, _writeTracebackMessage
from ..testing import (
    assertContainsFields,
    validateLogging,
    capture_logging,
    MemoryLogger,
)
from .._errors import register_exception_extractor
from .test_action import make_error_extraction_tests


def assert_expected_traceback(test, logger, message, exception, expected_traceback):
    """Assert we logged the given exception and the expected traceback."""
    lines = expected_traceback.split("\n")
    # Remove source code lines:
    expected_traceback = "\n".join([l for l in lines if not l.startswith("    ")])
    assertContainsFields(
        test,
        message,
        {
            "message_type": "eliot:traceback",
            "exception": RuntimeError,
            "reason": exception,
            "traceback": expected_traceback,
        },
    )
    logger.flushTracebacks(RuntimeError)


class TracebackLoggingTests(TestCase):
    """
    Tests for L{write_traceback} and L{writeFailure}.
    """

    @validateLogging(None)
    def test_write_traceback_implicit(self, logger):
        """
        L{write_traceback} with no arguments writes the current traceback to
        the log.
        """
        e = None

        def raiser():
            raise RuntimeError("because")

        try:
            raiser()
        except Exception as exception:
            expected_traceback = traceback.format_exc()
            write_traceback(logger)
            e = exception
        assert_expected_traceback(
            self, logger, logger.messages[0], e, expected_traceback
        )

    @validateLogging(None)
    def test_write_traceback_explicit(self, logger):
        """
        L{write_traceback} with explicit arguments writes the given traceback
        to the log.
        """
        e = None

        def raiser():
            raise RuntimeError("because")

        try:
            raiser()
        except Exception as exception:
            expected_traceback = traceback.format_exc()
            write_traceback(logger, exc_info=sys.exc_info())
            e = exception
        assert_expected_traceback(
            self, logger, logger.messages[0], e, expected_traceback
        )

    @capture_logging(None)
    def test_writeTracebackDefaultLogger(self, logger):
        """
        L{write_traceback} writes to the default log, if none is
        specified.
        """

        def raiser():
            raise RuntimeError("because")

        try:
            raiser()
        except Exception:
            write_traceback()

        message = logger.messages[0]
        assertContainsFields(self, message, {"message_type": "eliot:traceback"})
        logger.flushTracebacks(RuntimeError)

    @validateLogging(None)
    def test_writeFailure(self, logger):
        """
        L{writeFailure} writes a L{Failure} to the log.
        """
        if Failure is None:
            raise SkipTest("Twisted unavailable")

        try:
            raise RuntimeError("because")
        except:
            failure = Failure()
            expectedTraceback = failure.getBriefTraceback()
            writeFailure(failure, logger)
        message = logger.messages[0]
        assertContainsFields(
            self,
            message,
            {
                "message_type": "eliot:traceback",
                "exception": RuntimeError,
                "reason": failure.value,
                "traceback": expectedTraceback,
            },
        )
        logger.flushTracebacks(RuntimeError)

    @capture_logging(None)
    def test_writeFailureDefaultLogger(self, logger):
        """
        L{writeFailure} writes to the default log, if none is
        specified.
        """
        if Failure is None:
            raise SkipTest("Twisted unavailable")

        try:
            raise RuntimeError("because")
        except:
            failure = Failure()
            writeFailure(failure)
        message = logger.messages[0]
        assertContainsFields(self, message, {"message_type": "eliot:traceback"})
        logger.flushTracebacks(RuntimeError)

    @validateLogging(None)
    def test_writeFailureResult(self, logger):
        """
        L{writeFailure} returns C{None}.
        """
        if Failure is None:
            raise SkipTest("Twisted unavailable")

        try:
            raise RuntimeError("because")
        except:
            result = writeFailure(Failure(), logger)
        self.assertIs(result, None)
        logger.flushTracebacks(RuntimeError)

    @validateLogging(None)
    def test_serialization(self, logger):
        """
        L{_writeTracebackMessage} serializes exceptions to string values and
        types to FQPN.
        """
        try:
            raise KeyError(123)
        except:
            exc_info = sys.exc_info()
        _writeTracebackMessage(logger, *exc_info)
        serialized = logger.serialize()[0]
        assertContainsFields(
            self,
            serialized,
            {"exception": "%s.KeyError" % (KeyError.__module__,), "reason": "123"},
        )
        logger.flushTracebacks(KeyError)

    @validateLogging(None)
    def test_badException(self, logger):
        """
        L{_writeTracebackMessage} logs a message even if given a bad exception.
        """

        class BadException(Exception):
            def __str__(self):
                raise TypeError()

        try:
            raise BadException()
        except BadException:
            exc_info = sys.exc_info()
        _writeTracebackMessage(logger, *exc_info)
        self.assertEqual(
            logger.serialize()[0]["reason"],
            "eliot: unknown, unicode() raised exception",
        )
        logger.flushTracebacks(BadException)


def get_traceback_messages(exception):
    """
    Given an exception instance generate a traceback Eliot message.
    """
    logger = MemoryLogger()
    try:
        raise exception
    except exception.__class__:
        write_traceback(logger)
    # MemoryLogger.validate() mutates messages:
    # https://github.com/itamarst/eliot/issues/243
    messages = [message.copy() for message in logger.messages]
    logger.validate()
    return messages


class TracebackExtractionTests(make_error_extraction_tests(get_traceback_messages)):
    """
    Error extraction tests for tracebacks.
    """

    def test_regular_fields(self):
        """
        The normal traceback fields are still present when error
        extraction is used.
        """

        class MyException(Exception):
            pass

        register_exception_extractor(MyException, lambda e: {"key": e.args[0]})
        exception = MyException("because")
        messages = get_traceback_messages(exception)
        assertContainsFields(
            self,
            messages[0],
            {
                "message_type": "eliot:traceback",
                "reason": exception,
                "exception": MyException,
            },
        )



"""
Tests for L{eliot._output}.
"""

from sys import stdout
from unittest import TestCase, skipUnless

# Make sure to use StringIO that only accepts unicode:
from io import BytesIO, StringIO
import json as pyjson
from tempfile import mktemp
from time import time
from uuid import UUID
from threading import Thread

try:
    import numpy as np
except ImportError:
    np = None
from zope.interface.verify import verifyClass

from .._output import (
    MemoryLogger,
    ILogger,
    Destinations,
    Logger,
    bytesjson as json,
    to_file,
    FileDestination,
    _safe_unicode_dictionary,
)
from .._action import start_action
from .._validation import ValidationError, Field, _MessageSerializer
from .._traceback import write_traceback
from ..testing import assertContainsFields
from .common import CustomObject, CustomJSONEncoder


class MemoryLoggerTests(TestCase):
    """
    Tests for L{MemoryLogger}.
    """

    def test_interface(self):
        """
        L{MemoryLogger} implements L{ILogger}.
        """
        verifyClass(ILogger, MemoryLogger)

    def test_write(self):
        """
        Dictionaries written with L{MemoryLogger.write} are stored on a list.
        """
        logger = MemoryLogger()
        logger.write({"a": "b"})
        logger.write({"c": 1})
        self.assertEqual(logger.messages, [{"a": "b"}, {"c": 1}])
        logger.validate()

    def test_notStringFieldKeys(self):
        """
        Field keys must be unicode or bytes; if not L{MemoryLogger.validate}
        raises a C{TypeError}.
        """
        logger = MemoryLogger()
        logger.write({123: "b"})
        self.assertRaises(TypeError, logger.validate)

    def test_bytesMustBeUTF8(self):
        """
        Field keys can be bytes, but only if they're UTF-8 encoded Unicode.
        """
        logger = MemoryLogger()
        logger.write({"\u1234".encode("utf-16"): "b"})
        self.assertRaises(UnicodeDecodeError, logger.validate)

    def test_serializer(self):
        """
        L{MemoryLogger.validate} calls the given serializer's C{validate()}
        method with the message, as does L{MemoryLogger.write}.
        """

        class FakeValidator(list):
            def validate(self, message):
                self.append(message)

            def serialize(self, obj):
                return obj

        validator = FakeValidator()
        logger = MemoryLogger()
        message = {"message_type": "mymessage", "X": 1}
        logger.write(message, validator)
        self.assertEqual(validator, [message])
        logger.validate()
        self.assertEqual(validator, [message, message])

    def test_failedValidation(self):
        """
        L{MemoryLogger.validate} will allow exceptions raised by the serializer
        to pass through.
        """
        serializer = _MessageSerializer(
            [Field.forValue("message_type", "mymessage", "The type")]
        )
        logger = MemoryLogger()
        logger.write({"message_type": "wrongtype"}, serializer)
        self.assertRaises(ValidationError, logger.validate)

    def test_JSON(self):
        """
        L{MemoryLogger.validate} will encode the output of serialization to
        JSON.
        """
        serializer = _MessageSerializer(
            [
                Field.forValue("message_type", "type", "The type"),
                Field("foo", lambda value: object(), "The type"),
            ]
        )
        logger = MemoryLogger()
        logger.write(
            {"message_type": "type", "foo": "will become object()"}, serializer
        )
        self.assertRaises(TypeError, logger.validate)

    @skipUnless(np, "NumPy is not installed.")
    def test_EliotJSONEncoder(self):
        """
        L{MemoryLogger.validate} uses the EliotJSONEncoder by default to do
        encoding testing.
        """
        logger = MemoryLogger()
        logger.write({"message_type": "type", "foo": np.uint64(12)}, None)
        logger.validate()

    def test_JSON_custom_encoder(self):
        """
        L{MemoryLogger.validate} will use a custom JSON encoder if one was given.
        """
        logger = MemoryLogger(encoder=CustomJSONEncoder)
        logger.write(
            {"message_type": "type", "custom": CustomObject()},
            None,
        )
        logger.validate()

    def test_serialize(self):
        """
        L{MemoryLogger.serialize} returns a list of serialized versions of the
        logged messages.
        """
        serializer = _MessageSerializer(
            [
                Field.forValue("message_type", "mymessage", "The type"),
                Field("length", len, "The length"),
            ]
        )
        messages = [
            {"message_type": "mymessage", "length": "abc"},
            {"message_type": "mymessage", "length": "abcd"},
        ]
        logger = MemoryLogger()
        for message in messages:
            logger.write(message, serializer)
        self.assertEqual(
            logger.serialize(),
            [
                {"message_type": "mymessage", "length": 3},
                {"message_type": "mymessage", "length": 4},
            ],
        )

    def test_serializeCopies(self):
        """
        L{MemoryLogger.serialize} does not mutate the original logged messages.
        """
        serializer = _MessageSerializer(
            [
                Field.forValue("message_type", "mymessage", "The type"),
                Field("length", len, "The length"),
            ]
        )
        message = {"message_type": "mymessage", "length": "abc"}
        logger = MemoryLogger()
        logger.write(message, serializer)
        logger.serialize()
        self.assertEqual(logger.messages[0]["length"], "abc")

    def write_traceback(self, logger, exception):
        """
        Write an exception as a traceback to the logger.
        """
        try:
            raise exception
        except:
            write_traceback(logger)

    def test_tracebacksCauseTestFailure(self):
        """
        Logging a traceback to L{MemoryLogger} will add its exception to
        L{MemoryLogger.tracebackMessages}.
        """
        logger = MemoryLogger()
        exception = Exception()
        self.write_traceback(logger, exception)
        self.assertEqual(logger.tracebackMessages[0]["reason"], exception)

    def test_flushTracebacksNoTestFailure(self):
        """
        Any tracebacks cleared by L{MemoryLogger.flushTracebacks} (as specified
        by exception type) are removed from
        L{MemoryLogger.tracebackMessages}.
        """
        logger = MemoryLogger()
        exception = RuntimeError()
        self.write_traceback(logger, exception)
        logger.flushTracebacks(RuntimeError)
        self.assertEqual(logger.tracebackMessages, [])

    def test_flushTracebacksReturnsExceptions(self):
        """
        L{MemoryLogger.flushTracebacks} returns the traceback messages.
        """
        exceptions = [ZeroDivisionError(), ZeroDivisionError()]
        logger = MemoryLogger()
        logger.write({"x": 1})
        for exc in exceptions:
            self.write_traceback(logger, exc)
        logger.write({"x": 1})
        flushed = logger.flushTracebacks(ZeroDivisionError)
        self.assertEqual(flushed, logger.messages[1:3])

    def test_flushTracebacksUnflushedTestFailure(self):
        """
        Any tracebacks uncleared by L{MemoryLogger.flushTracebacks} (because
        they are of a different type) are still listed in
        L{MemoryLogger.tracebackMessages}.
        """
        logger = MemoryLogger()
        exception = RuntimeError()
        self.write_traceback(logger, exception)
        logger.flushTracebacks(KeyError)
        self.assertEqual(logger.tracebackMessages[0]["reason"], exception)

    def test_flushTracebacksUnflushedUnreturned(self):
        """
        Any tracebacks uncleared by L{MemoryLogger.flushTracebacks} (because
        they are of a different type) are not returned.
        """
        logger = MemoryLogger()
        exception = RuntimeError()
        self.write_traceback(logger, exception)
        self.assertEqual(logger.flushTracebacks(KeyError), [])

    def test_reset(self):
        """
        L{MemoryLogger.reset} clears all logged messages and tracebacks.
        """
        logger = MemoryLogger()
        logger.write({"key": "value"}, None)
        logger.reset()
        self.assertEqual(
            (logger.messages, logger.serializers, logger.tracebackMessages),
            ([], [], []),
        )

    def test_threadSafeWrite(self):
        """
        L{MemoryLogger.write} can be called from multiple threads concurrently.
        """
        # Some threads will log some messages
        thread_count = 10

        # A lot of messages.  This will keep the threads running long enough
        # to give them a chance to (try to) interfere with each other.
        write_count = 10000

        # They'll all use the same MemoryLogger instance.
        logger = MemoryLogger()

        # Each thread will have its own message and serializer that it writes
        # to the log over and over again.
        def write(msg, serializer):
            for i in range(write_count):
                logger.write(msg, serializer)

        # Generate a single distinct message for each thread to log.
        msgs = list({"i": i} for i in range(thread_count))

        # Generate a single distinct serializer for each thread to log.
        serializers = list(object() for i in range(thread_count))

        # Pair them all up.  This gives us a simple invariant we can check
        # later on.
        write_args = zip(msgs, serializers)

        # Create the threads.
        threads = list(Thread(target=write, args=args) for args in write_args)

        # Run them all.  Note threads early in this list will start writing to
        # the log before later threads in the list even get a chance to start.
        # That's part of why we have each thread write so many messages.
        for t in threads:
            t.start()
        # Wait for them all to finish.
        for t in threads:
            t.join()

        # Check that we got the correct number of messages in the log.
        expected_count = thread_count * write_count
        self.assertEqual(len(logger.messages), expected_count)
        self.assertEqual(len(logger.serializers), expected_count)

        # Check the simple invariant we created above.  Every logged message
        # must be paired with the correct serializer, where "correct" is
        # defined by ``write_args`` above.
        for position, (msg, serializer) in enumerate(
            zip(logger.messages, logger.serializers)
        ):
            # The indexes must match because the objects are paired using
            # zip() above.
            msg_index = msgs.index(msg)
            serializer_index = serializers.index(serializer)
            self.assertEqual(
                msg_index,
                serializer_index,
                "Found message #{} with serializer #{} at position {}".format(
                    msg_index, serializer_index, position
                ),
            )


class MyException(Exception):
    """
    Custom exception.
    """


class BadDestination(list):
    """
    A destination that throws an exception the first time it is called.
    """

    called = 0

    def __call__(self, msg):
        if not self.called:
            self.called = True
            raise MyException("ono")
        self.append(msg)


class DestinationsTests(TestCase):
    """
    Tests for L{Destinations}.
    """

    def test_send(self):
        """
        L{Destinations.send} calls all destinations added with
        L{Destinations.add} with the given dictionary.
        """
        destinations = Destinations()
        message = {"hoorj": "blargh"}
        dest = []
        dest2 = []
        dest3 = []
        destinations.add(dest.append, dest2.append)
        destinations.add(dest3.append)
        destinations.send(message)
        self.assertEqual(dest, [message])
        self.assertEqual(dest2, [message])
        self.assertEqual(dest3, [message])

    def test_destination_exception_multiple_destinations(self):
        """
        If one destination throws an exception, other destinations still
        get the message.
        """
        destinations = Destinations()
        dest = []
        dest2 = BadDestination()
        dest3 = []
        destinations.add(dest.append)
        destinations.add(dest2)
        destinations.add(dest3.append)

        message = {"hello": 123}
        destinations.send(message)
        self.assertIn(message, dest)
        self.assertIn(message, dest3)

    def test_destination_exception_continue(self):
        """
        If a destination throws an exception, future messages are still
        sent to it.
        """
        destinations = Destinations()
        dest = BadDestination()
        destinations.add(dest)

        msg1 = {"hello": 123}
        msg2 = {"world": 456}
        destinations.send(msg1)
        self.assertNotIn(msg1, dest)
        destinations.send(msg2)
        self.assertIn(msg2, dest)

    def test_remove(self):
        """
        A destination removed with L{Destinations.remove} will no longer
        receive messages from L{Destionations.add} calls.
        """
        destinations = Destinations()
        message = {"hello": 123}
        dest = []
        destinations.add(dest.append)
        destinations.remove(dest.append)
        destinations.send(message)
        self.assertEqual(dest, [])

    def test_removeNonExistent(self):
        """
        Removing a destination that has not previously been added with result
        in a C{ValueError} being thrown.
        """
        destinations = Destinations()
        self.assertRaises(ValueError, destinations.remove, [].append)

    def test_addGlobalFields(self):
        """
        L{Destinations.addGlobalFields} adds the given fields and values to
        the messages being passed in.
        """
        destinations = Destinations()
        dest = []
        destinations.add(dest.append)
        destinations.addGlobalFields(x=123, y="hello")
        destinations.send({"z": 456})
        self.assertEqual(dest, [{"x": 123, "y": "hello", "z": 456}])

    def test_addGlobalFieldsCumulative(self):
        """
        L{Destinations.addGlobalFields} adds the given fields to those set by
        previous calls.
        """
        destinations = Destinations()
        dest = []
        destinations.add(dest.append)
        destinations.addGlobalFields(x=123, y="hello")
        destinations.addGlobalFields(x=456, z=456)
        destinations.send({"msg": "X"})
        self.assertEqual(dest, [{"x": 456, "y": "hello", "z": 456, "msg": "X"}])

    def test_buffering(self):
        """
        Before any destinations are set up to 1000 messages are buffered, and
        then delivered to the first registered destinations.
        """
        destinations = Destinations()
        messages = [{"k": i} for i in range(1050)]
        for m in messages:
            destinations.send(m)
        dest, dest2 = [], []
        destinations.add(dest.append, dest2.append)
        self.assertEqual((dest, dest2), (messages[-1000:], messages[-1000:]))

    def test_buffering_second_batch(self):
        """
        The second batch of added destination don't get the buffered messages.
        """
        destinations = Destinations()
        message = {"m": 1}
        message2 = {"m": 2}
        destinations.send(message)
        dest = []
        dest2 = []
        destinations.add(dest.append)
        destinations.add(dest2.append)
        destinations.send(message2)
        self.assertEqual((dest, dest2), ([message, message2], [message2]))

    def test_global_fields_buffering(self):
        """
        Global fields are added to buffered messages, when possible.
        """
        destinations = Destinations()
        message = {"m": 1}
        destinations.send(message)
        destinations.addGlobalFields(k=123)
        dest = []
        destinations.add(dest.append)
        self.assertEqual(dest, [{"m": 1, "k": 123}])


def makeLogger():
    """
    Return a tuple (L{Logger} instance, C{list} of written messages).
    """
    logger = Logger()
    logger._destinations = Destinations()
    written = []
    logger._destinations.add(written.append)
    return logger, written


class LoggerTests(TestCase):
    """
    Tests for L{Logger}.
    """

    def test_interface(self):
        """
        L{Logger} implements L{ILogger}.
        """
        verifyClass(ILogger, Logger)

    def test_global(self):
        """
        A global L{Destinations} is used by the L{Logger} class.
        """
        self.assertIsInstance(Logger._destinations, Destinations)

    def test_write(self):
        """
        L{Logger.write} sends the given dictionary L{Destinations} object.
        """
        logger, written = makeLogger()

        d = {"hello": 1}
        logger.write(d)
        self.assertEqual(written, [d])

    def test_serializer(self):
        """
        If a L{_MessageSerializer} is passed to L{Logger.write}, it is used to
        serialize the message before it is passed to the destination.
        """
        logger, written = makeLogger()

        serializer = _MessageSerializer(
            [
                Field.forValue("message_type", "mymessage", "The type"),
                Field("length", len, "The length of a thing"),
            ]
        )
        logger.write({"message_type": "mymessage", "length": "thething"}, serializer)
        self.assertEqual(written, [{"message_type": "mymessage", "length": 8}])

    def test_passedInDictionaryUnmodified(self):
        """
        The dictionary passed in to L{Logger.write} is not modified.
        """
        logger, written = makeLogger()

        serializer = _MessageSerializer(
            [
                Field.forValue("message_type", "mymessage", "The type"),
                Field("length", len, "The length of a thing"),
            ]
        )
        d = {"message_type": "mymessage", "length": "thething"}
        original = d.copy()
        logger.write(d, serializer)
        self.assertEqual(d, original)

    def test_safe_unicode_dictionary(self):
        """
        L{_safe_unicode_dictionary} converts the given dictionary's
        values and keys to unicode using C{safeunicode}.
        """

        class badobject(object):
            def __repr__(self):
                raise TypeError()

        dictionary = {badobject(): 123, 123: badobject()}
        badMessage = "eliot: unknown, unicode() raised exception"
        self.assertEqual(
            eval(_safe_unicode_dictionary(dictionary)),
            {badMessage: "123", "123": badMessage},
        )

    def test_safe_unicode_dictionary_fallback(self):
        """
        If converting the dictionary failed for some reason,
        L{_safe_unicode_dictionary} runs C{repr} on the object.
        """
        self.assertEqual(_safe_unicode_dictionary(None), "None")

    def test_safe_unicode_dictionary_fallback_failure(self):
        """
        If all else fails, L{_safe_unicode_dictionary} just gives up.
        """

        class badobject(object):
            def __repr__(self):
                raise TypeError()

        self.assertEqual(
            _safe_unicode_dictionary(badobject()),
            "eliot: unknown, unicode() raised exception",
        )

    def test_serializationErrorTraceback(self):
        """
        If serialization fails in L{Logger.write}, a traceback is logged,
        along with a C{eliot:serialization_failure} message for debugging
        purposes.
        """
        logger, written = makeLogger()

        def raiser(i):
            raise RuntimeError("oops")

        serializer = _MessageSerializer(
            [
                Field.forValue("message_type", "mymessage", "The type"),
                Field("fail", raiser, "Serialization fail"),
            ]
        )
        message = {"message_type": "mymessage", "fail": "will"}
        logger.write(message, serializer)
        self.assertEqual(len(written), 2)
        tracebackMessage = written[0]
        assertContainsFields(
            self,
            tracebackMessage,
            {
                "exception": "%s.RuntimeError" % (RuntimeError.__module__,),
                "message_type": "eliot:traceback",
            },
        )
        self.assertIn("RuntimeError: oops", tracebackMessage["traceback"])
        # Calling _safe_unicode_dictionary multiple times leads to
        # inconsistent results due to hash ordering, so compare contents:
        assertContainsFields(
            self, written[1], {"message_type": "eliot:serialization_failure"}
        )
        self.assertEqual(
            eval(written[1]["message"]),
            dict((repr(key), repr(value)) for (key, value) in message.items()),
        )

    def test_destination_exception_caught(self):
        """
        If a destination throws an exception, an appropriate error is
        logged.
        """
        logger = Logger()
        logger._destinations = Destinations()
        dest = BadDestination()
        logger._destinations.add(dest)

        message = {"hello": 123}
        logger.write({"hello": 123})
        assertContainsFields(
            self,
            dest[0],
            {
                "message_type": "eliot:destination_failure",
                "message": _safe_unicode_dictionary(message),
                "reason": "ono",
                "exception": "eliot.tests.test_output.MyException",
            },
        )

    def test_destination_multiple_exceptions_caught(self):
        """
        If multiple destinations throw an exception, an appropriate error is
        logged for each.
        """
        logger = Logger()
        logger._destinations = Destinations()
        logger._destinations.add(BadDestination())
        logger._destinations.add(lambda msg: 1 / 0)
        messages = []
        logger._destinations.add(messages.append)

        try:
            1 / 0
        except ZeroDivisionError as e:
            zero_divide = str(e)
        zero_type = ZeroDivisionError.__module__ + ".ZeroDivisionError"

        message = {"hello": 123}
        logger.write({"hello": 123})

        def remove(key):
            return [message.pop(key) for message in messages[1:]]

        # Make sure we have task_level & task_uuid in exception messages.
        task_levels = remove("task_level")
        task_uuids = remove("task_uuid")
        timestamps = remove("timestamp")

        self.assertEqual(
            (
                abs(timestamps[0] + timestamps[1] - 2 * time()) < 1,
                task_levels == [[1], [1]],
                len([UUID(uuid) for uuid in task_uuids]) == 2,
                messages,
            ),
            (
                True,
                True,
                True,
                [
                    message,
                    {
                        "message_type": "eliot:destination_failure",
                        "message": _safe_unicode_dictionary(message),
                        "reason": "ono",
                        "exception": "eliot.tests.test_output.MyException",
                    },
                    {
                        "message_type": "eliot:destination_failure",
                        "message": _safe_unicode_dictionary(message),
                        "reason": zero_divide,
                        "exception": zero_type,
                    },
                ],
            ),
        )

    def test_destination_exception_caught_twice(self):
        """
        If a destination throws an exception, and the logged error about
        it also causes an exception, then just drop that exception on the
        floor, since there's nothing we can do with it.
        """
        logger = Logger()
        logger._destinations = Destinations()

        def always_raise(message):
            raise ZeroDivisionError()

        logger._destinations.add(always_raise)

        # Just a message. No exception raised; since everything is dropped no
        # other assertions to be made.
        logger.write({"hello": 123})

        # With an action. No exception raised; since everything is dropped no
        # other assertions to be made.
        with start_action(logger, "sys:do"):
            logger.write({"hello": 123})


class PEP8Tests(TestCase):
    """
    Tests for PEP 8 method compatibility.
    """

    def test_flush_tracebacks(self):
        """
        L{MemoryLogger.flush_tracebacks} is the same as
        L{MemoryLogger.flushTracebacks}
        """
        self.assertEqual(MemoryLogger.flush_tracebacks, MemoryLogger.flushTracebacks)


class ToFileTests(TestCase):
    """
    Tests for L{to_file}.
    """

    def test_to_file_adds_destination(self):
        """
        L{to_file} adds a L{FileDestination} destination with the given file.
        """
        f = stdout
        to_file(f)
        expected = FileDestination(file=f)
        self.addCleanup(Logger._destinations.remove, expected)
        self.assertIn(expected, Logger._destinations._destinations)

    def test_to_file_custom_encoder(self):
        """
        L{to_file} accepts a custom encoder, and sets it on the resulting
        L{FileDestination}.
        """
        f = stdout
        encoder = object()
        to_file(f, encoder=encoder)
        expected = FileDestination(file=f, encoder=encoder)
        self.addCleanup(Logger._destinations.remove, expected)
        self.assertIn(expected, Logger._destinations._destinations)

    def test_bytes_values(self):
        """
        DEPRECATED: On Python 3L{FileDestination} will encode bytes as if they were
        UTF-8 encoded strings when writing to BytesIO only.
        """
        message = {"x": b"abc"}
        bytes_f = BytesIO()
        destination = FileDestination(file=bytes_f)
        destination(message)
        self.assertEqual(
            [json.loads(line) for line in bytes_f.getvalue().splitlines()],
            [{"x": "abc"}],
        )

    @skipUnless(np, "NumPy is not installed.")
    def test_default_encoder_is_EliotJSONEncoder(self):
        """The default encoder if none are specified is EliotJSONEncoder."""
        message = {"x": np.int64(3)}
        f = StringIO()
        destination = FileDestination(file=f)
        destination(message)
        self.assertEqual(
            [json.loads(line) for line in f.getvalue().splitlines()], [{"x": 3}]
        )

    def test_filedestination_writes_json_bytes(self):
        """
        L{FileDestination} writes JSON-encoded messages to a file that accepts
        bytes.
        """
        message1 = {"x": 123}
        message2 = {"y": None, "x": "abc"}
        bytes_f = BytesIO()
        destination = FileDestination(file=bytes_f)
        destination(message1)
        destination(message2)
        self.assertEqual(
            [json.loads(line) for line in bytes_f.getvalue().splitlines()],
            [message1, message2],
        )

    def test_filedestination_custom_encoder(self):
        """
        L{FileDestionation} can use a custom encoder.
        """
        custom = object()

        class CustomEncoder(pyjson.JSONEncoder):
            def default(self, o):
                if o is custom:
                    return "CUSTOM!"
                else:
                    return pyjson.JSONEncoder.default(self, o)

        message = {"x": 123, "z": custom}
        f = BytesIO()
        destination = FileDestination(file=f, encoder=CustomEncoder)
        destination(message)
        self.assertEqual(
            json.loads(f.getvalue().splitlines()[0]), {"x": 123, "z": "CUSTOM!"}
        )

    def test_filedestination_flushes(self):
        """
        L{FileDestination} flushes after every write, to ensure logs get
        written out even if the local buffer hasn't filled up.
        """
        path = mktemp()
        # File with large buffer:
        f = open(path, "wb", 1024 * 1024 * 10)
        # and a small message that won't fill the buffer:
        message1 = {"x": 123}

        destination = FileDestination(file=f)
        destination(message1)

        # Message got written even though buffer wasn't filled:
        self.assertEqual(
            [json.loads(line) for line in open(path, "rb").read().splitlines()],
            [message1],
        )

    def test_filedestination_writes_json_unicode(self):
        """
        L{FileDestination} writes JSON-encoded messages to file that only
        accepts Unicode.
        """
        message = {"x": "\u1234"}
        unicode_f = StringIO()
        destination = FileDestination(file=unicode_f)
        destination(message)
        self.assertEqual(pyjson.loads(unicode_f.getvalue()), message)

    def test_filedestination_unwriteable_file(self):
        """
        L{FileDestination} raises a runtime error if the given file isn't writeable.
        """
        path = mktemp()
        open(path, "w").close()
        f = open(path, "r")
        with self.assertRaises(RuntimeError):
            FileDestination(f)



"""Support for Eliot tracing with Dask computations."""

from pyrsistent import PClass, field

from dask import compute, optimize, persist

try:
    from dask.distributed import Future
except:

    class Future(object):
        pass


from dask.core import toposort, get_dependencies, ishashable
from . import start_action, current_action, Action


class _RunWithEliotContext(PClass):
    """
    Run a callable within an Eliot context.

    @ivar task_id: The serialized Eliot task ID.
    @ivar func: The function that Dask wants to run.
    @ivar key: The key in the Dask graph.
    @ivar dependencies: The keys in the Dask graph this depends on.
    """

    task_id = field(type=str)
    func = field()  # callable
    key = field(type=str)
    dependencies = field()

    # Pretend to be underlying callable for purposes of equality; necessary for
    # optimizer to be happy:

    def __eq__(self, other):
        return self.func == other

    def __ne__(self, other):
        return self.func != other

    def __hash__(self):
        return hash(self.func)

    def __call__(self, *args, **kwargs):
        with Action.continue_task(task_id=self.task_id) as action:
            action.log(
                message_type="dask:task", key=self.key, dependencies=self.dependencies
            )
            return self.func(*args, **kwargs)


def compute_with_trace(*args):
    """Do Dask compute(), but with added Eliot tracing.

    Dask is a graph of tasks, but Eliot logs trees.  So we need to emulate a
    graph using a tree.  We do this by making Eliot action for each task, but
    having it list the tasks it depends on.

    We use the following algorithm:

        1. Create a top-level action.

        2. For each entry in the dask graph, create a child with
           serialize_task_id.  Do this in likely order of execution, so that
           if B depends on A the task level of B is higher than the task Ievel
           of A.

        3. Replace each function with a wrapper that uses the corresponding
           task ID (with Action.continue_task), and while it's at it also
           records which other things this function depends on.

    Known issues:

        1. Retries will confuse Eliot.  Probably need different
           distributed-tree mechanism within Eliot to solve that.
    """
    # 1. Create top-level Eliot Action:
    with start_action(action_type="dask:compute"):
        # In order to reduce logging verbosity, add logging to the already
        # optimized graph:
        optimized = optimize(*args, optimizations=[_add_logging])
        return compute(*optimized, optimize_graph=False)


def persist_with_trace(*args):
    """Do Dask persist(), but with added Eliot tracing.

    Known issues:

        1. Retries will confuse Eliot.  Probably need different
           distributed-tree mechanism within Eliot to solve that.
    """
    # 1. Create top-level Eliot Action:
    with start_action(action_type="dask:persist"):
        # In order to reduce logging verbosity, add logging to the already
        # optimized graph:
        optimized = optimize(*args, optimizations=[_add_logging])
        return persist(*optimized, optimize_graph=False)


def _add_logging(dsk, ignore=None):
    """
    Add logging to a Dask graph.

    @param dsk: The Dask graph.

    @return: New Dask graph.
    """
    ctx = current_action()
    result = {}

    # Use topological sort to ensure Eliot actions are in logical order of
    # execution in Dask:
    keys = toposort(dsk)

    # Give each key a string name. Some keys are just aliases to other
    # keys, so make sure we have underlying key available. Later on might
    # want to shorten them as well.
    def simplify(k):
        if isinstance(k, str):
            return k
        return "-".join(str(o) for o in k)

    key_names = {}
    for key in keys:
        value = dsk[key]
        if not callable(value) and ishashable(value) and value in keys:
            # It's an alias for another key:
            key_names[key] = key_names[value]
        else:
            key_names[key] = simplify(key)

    # Values in the graph can be either:
    #
    # 1. A list of other values.
    # 2. A tuple, where first value might be a callable, aka a task.
    # 3. A literal of some sort.
    def maybe_wrap(key, value):
        if isinstance(value, list):
            return [maybe_wrap(key, v) for v in value]
        elif isinstance(value, tuple):
            func = value[0]
            args = value[1:]
            if not callable(func):
                # Not a callable, so nothing to wrap.
                return value
            wrapped_func = _RunWithEliotContext(
                task_id=str(ctx.serialize_task_id(), "utf-8"),
                func=func,
                key=key_names[key],
                dependencies=[key_names[k] for k in get_dependencies(dsk, key)],
            )
            return (wrapped_func,) + args
        else:
            return value

    # Replace function with wrapper that logs appropriate Action; iterate in
    # topological order so action task levels are in reasonable order.
    for key in keys:
        result[key] = maybe_wrap(key, dsk[key])

    assert set(result.keys()) == set(dsk.keys())
    return result


__all__ = ["compute_with_trace", "persist_with_trace"]



"""
API and command-line support for human-readable Eliot messages.
"""

import pprint
import argparse
from datetime import datetime
from sys import stdin, stdout
from collections import OrderedDict
from json import dumps

from ._bytesjson import loads
from ._message import (
    TIMESTAMP_FIELD,
    TASK_UUID_FIELD,
    TASK_LEVEL_FIELD,
    MESSAGE_TYPE_FIELD,
)
from ._action import ACTION_TYPE_FIELD, ACTION_STATUS_FIELD


# Ensure binary stdin, since we expect specifically UTF-8 encoded
# messages, not platform-encoding messages.
stdin = stdin.buffer


# Fields that all Eliot messages are expected to have:
REQUIRED_FIELDS = {TASK_LEVEL_FIELD, TASK_UUID_FIELD, TIMESTAMP_FIELD}

# Fields that get treated specially when formatting.
_skip_fields = {
    TIMESTAMP_FIELD,
    TASK_UUID_FIELD,
    TASK_LEVEL_FIELD,
    MESSAGE_TYPE_FIELD,
    ACTION_TYPE_FIELD,
    ACTION_STATUS_FIELD,
}

# First fields to render:
_first_fields = [ACTION_TYPE_FIELD, MESSAGE_TYPE_FIELD, ACTION_STATUS_FIELD]


def _render_timestamp(message: dict, local_timezone: bool) -> str:
    """Convert a message's timestamp to a string."""
    # If we were returning or storing the datetime we'd want to use an
    # explicit timezone instead of a naive datetime, but since we're
    # just using it for formatting we needn't bother.
    if local_timezone:
        dt = datetime.fromtimestamp(message[TIMESTAMP_FIELD])
    else:
        dt = datetime.utcfromtimestamp(message[TIMESTAMP_FIELD])
    result = dt.isoformat(sep="T")
    if not local_timezone:
        result += "Z"
    return result


def pretty_format(message: dict, local_timezone: bool = False) -> str:
    """
    Convert a message dictionary into a human-readable string.

    @param message: Message to parse, as dictionary.

    @return: Unicode string.
    """

    def add_field(previous, key, value):
        value = (
            pprint.pformat(value, width=40).replace("\\n", "\n ").replace("\\t", "\t")
        )
        # Reindent second line and later to match up with first line's
        # indentation:
        lines = value.split("\n")
        # indent lines are "  <key length>|  <value>"
        indent = "{}| ".format(" " * (2 + len(key)))
        value = "\n".join([lines[0]] + [indent + l for l in lines[1:]])
        return "  %s: %s\n" % (key, value)

    remaining = ""
    for field in _first_fields:
        if field in message:
            remaining += add_field(remaining, field, message[field])
    for (key, value) in sorted(message.items()):
        if key not in _skip_fields:
            remaining += add_field(remaining, key, value)

    level = "/" + "/".join(map(str, message[TASK_LEVEL_FIELD]))
    return "%s -> %s\n%s\n%s" % (
        message[TASK_UUID_FIELD],
        level,
        _render_timestamp(message, local_timezone),
        remaining,
    )


def compact_format(message: dict, local_timezone: bool = False) -> str:
    """Format an Eliot message into a single line.

    The message is presumed to be JSON-serializable.
    """
    ordered_message = OrderedDict()
    for field in _first_fields:
        if field in message:
            ordered_message[field] = message[field]
    for (key, value) in sorted(message.items()):
        if key not in _skip_fields:
            ordered_message[key] = value
    # drop { and } from JSON:
    rendered = " ".join(
        "{}={}".format(key, dumps(value, separators=(",", ":")))
        for (key, value) in ordered_message.items()
    )

    return "%s%s %s %s" % (
        message[TASK_UUID_FIELD],
        "/" + "/".join(map(str, message[TASK_LEVEL_FIELD])),
        _render_timestamp(message, local_timezone),
        rendered,
    )


_CLI_HELP = """\
Convert Eliot messages into more readable format.

Reads JSON lines from stdin, write out pretty-printed results on stdout.
"""


def _main():
    """
    Command-line program that reads in JSON from stdin and writes out
    pretty-printed messages to stdout.
    """
    parser = argparse.ArgumentParser(
        description=_CLI_HELP, usage="cat messages | %(prog)s [options]"
    )
    parser.add_argument(
        "-c",
        "--compact",
        action="store_true",
        dest="compact",
        help="Compact format, one message per line.",
    )
    parser.add_argument(
        "-l",
        "--local-timezone",
        action="store_true",
        dest="local_timezone",
        help="Use local timezone instead of UTC.",
    )

    args = parser.parse_args()
    if args.compact:
        formatter = compact_format
    else:
        formatter = pretty_format

    for line in stdin:
        try:
            message = loads(line)
        except ValueError:
            stdout.write("Not JSON: {}\n\n".format(line.rstrip(b"\n")))
            continue
        if REQUIRED_FIELDS - set(message.keys()):
            stdout.write("Not an Eliot message: {}\n\n".format(line.rstrip(b"\n")))
            continue
        result = formatter(message, args.local_timezone) + "\n"
        stdout.write(result)


__all__ = ["pretty_format", "compact_format"]



"""
Log messages and related utilities.
"""

from __future__ import unicode_literals

import time
from warnings import warn
from six import text_type as unicode

from pyrsistent import PClass, pmap_field

MESSAGE_TYPE_FIELD = "message_type"
TASK_UUID_FIELD = "task_uuid"
TASK_LEVEL_FIELD = "task_level"
TIMESTAMP_FIELD = "timestamp"

EXCEPTION_FIELD = "exception"
REASON_FIELD = "reason"


class Message(object):
    """
    A log message.

    Messages are basically dictionaries, mapping "fields" to "values". Field
    names should not start with C{'_'}, as those are reserved for system use
    (e.g. C{"_id"} is used by Elasticsearch for unique message identifiers and
    may be auto-populated by logstash).
    """

    # Overrideable for testing purposes:
    _time = time.time

    @classmethod
    def new(_class, _serializer=None, **fields):
        """
        Create a new L{Message}.

        The keyword arguments will become the initial contents of the L{Message}.

        @param _serializer: A positional argument, either C{None} or a
            L{eliot._validation._MessageSerializer} with which a
            L{eliot.ILogger} may choose to serialize the message. If you're
            using L{eliot.MessageType} this will be populated for you.

        @return: The new L{Message}
        """
        warn(
            "Message.new() is deprecated since 1.11.0, "
            "use eliot.log_message() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return _class(fields, _serializer)

    @classmethod
    def log(_class, **fields):
        """
        Write a new L{Message} to the default L{Logger}.

        The keyword arguments will become contents of the L{Message}.
        """
        warn(
            "Message.log() is deprecated since 1.11.0, "
            "use Action.log() or eliot.log_message() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        _class(fields).write()

    def __init__(self, contents, serializer=None):
        """
        You can also use L{Message.new} to create L{Message} objects.

        @param contents: The contents of this L{Message}, a C{dict} whose keys
           must be C{unicode}, or text that has been UTF-8 encoded to
           C{bytes}.

        @param serializer: Either C{None}, or
            L{eliot._validation._MessageSerializer} with which a
            L{eliot.Logger} may choose to serialize the message. If you're
            using L{eliot.MessageType} this will be populated for you.
        """
        self._contents = contents.copy()
        self._serializer = serializer

    def bind(self, **fields):
        """
        Return a new L{Message} with this message's contents plus the
        additional given bindings.
        """
        contents = self._contents.copy()
        contents.update(fields)
        return Message(contents, self._serializer)

    def contents(self):
        """
        Return a copy of L{Message} contents.
        """
        return self._contents.copy()

    def _timestamp(self):
        """
        Return the current time.
        """
        return self._time()

    def write(self, logger=None, action=None):
        """
        Write the message to the given logger.

        This will additionally include a timestamp, the action context if any,
        and any other fields.

        Byte field names will be converted to Unicode.

        @type logger: L{eliot.ILogger} or C{None} indicating the default one.
            Should not be set if the action is also set.

        @param action: The L{Action} which is the context for this message.  If
            C{None}, the L{Action} will be deduced from the current call stack.
        """
        fields = dict(self._contents)
        if "message_type" not in fields:
            fields["message_type"] = ""
        if self._serializer is not None:
            fields["__eliot_serializer__"] = self._serializer
        if action is None:
            if logger is not None:
                fields["__eliot_logger__"] = logger
            log_message(**fields)
        else:
            action.log(**fields)


class WrittenMessage(PClass):
    """
    A L{Message} that has been logged.

    @ivar _logged_dict: The originally logged dictionary.
    """

    _logged_dict = pmap_field((str, unicode), object)

    @property
    def timestamp(self):
        """
        The Unix timestamp of when the message was logged.
        """
        return self._logged_dict[TIMESTAMP_FIELD]

    @property
    def task_uuid(self):
        """
        The UUID of the task in which the message was logged.
        """
        return self._logged_dict[TASK_UUID_FIELD]

    @property
    def task_level(self):
        """
        The L{TaskLevel} of this message appears within the task.
        """
        return TaskLevel(level=self._logged_dict[TASK_LEVEL_FIELD])

    @property
    def contents(self):
        """
        A C{PMap}, the message contents without Eliot metadata.
        """
        return (
            self._logged_dict.discard(TIMESTAMP_FIELD)
            .discard(TASK_UUID_FIELD)
            .discard(TASK_LEVEL_FIELD)
        )

    @classmethod
    def from_dict(cls, logged_dictionary):
        """
        Reconstruct a L{WrittenMessage} from a logged dictionary.

        @param logged_dictionary: A C{PMap} representing a parsed log entry.
        @return: A L{WrittenMessage} for that dictionary.
        """
        return cls(_logged_dict=logged_dictionary)

    def as_dict(self):
        """
        Return the dictionary that was used to write this message.

        @return: A C{dict}, as might be logged by Eliot.
        """
        return self._logged_dict


# Import at end to deal with circular imports:
from ._action import log_message, TaskLevel



"""
Tests for L{eliot._util}.
"""

from __future__ import unicode_literals

from unittest import TestCase
import pprint

from .._util import load_module


class LoadModuleTests(TestCase):
    """
    Tests for L{load_module}.
    """

    maxDiff = None

    def test_returns_module(self):
        """
        L{load_module} returns an object with same methods as original module.
        """
        loaded = load_module(str("copy"), pprint)
        obj = [1, 2, b"hello"]
        self.assertEqual(loaded.pformat(obj), pprint.pformat(obj))

    def test_name(self):
        """
        L{load_module} returns an object with the given name.
        """
        name = str("my_copy")
        loaded = load_module(name, pprint)
        self.assertEqual(loaded.__name__, name)

    def test_distinct_from_original(self):
        """
        L{load_module} returns a distinct object from the original module.
        """
        loaded = load_module(str("copy"), pprint)
        # Override repr in copy:
        loaded.repr = lambda o: str("OVERRIDE")
        # Demonstrate that override applies to copy but not original:
        self.assertEqual(
            dict(original=pprint.pformat(123), loaded=loaded.pformat(123)),
            dict(original="123", loaded="OVERRIDE"),
        )



"""
TAI64N encoding and decoding.

TAI64N encodes nanosecond-accuracy timestamps and is supported by logstash.

@see: U{http://cr.yp.to/libtai/tai64.html}.
"""

from __future__ import unicode_literals

import struct
from binascii import b2a_hex, a2b_hex

_STRUCTURE = b">QI"
_OFFSET = (2 ** 62) + 10  # last 10 are leap seconds


def encode(timestamp):
    """
    Convert seconds since epoch to TAI64N string.

    @param timestamp: Seconds since UTC Unix epoch as C{float}.

    @return: TAI64N-encoded time, as C{unicode}.
    """
    seconds = int(timestamp)
    nanoseconds = int((timestamp - seconds) * 1000000000)
    seconds = seconds + _OFFSET
    encoded = b2a_hex(struct.pack(_STRUCTURE, seconds, nanoseconds))
    return "@" + encoded.decode("ascii")


def decode(tai64n):
    """
    Convert TAI64N string to seconds since epoch.

    Note that dates before 2013 may not decode accurately due to leap second
    issues. If you need correct decoding for earlier dates you can try the
    tai64n package available from PyPI (U{https://pypi.python.org/pypi/tai64n}).

    @param tai64n: TAI64N-encoded time, as C{unicode}.

    @return: Seconds since UTC Unix epoch as C{float}.
    """
    seconds, nanoseconds = struct.unpack(_STRUCTURE, a2b_hex(tai64n[1:]))
    seconds -= _OFFSET
    return seconds + (nanoseconds / 1000000000.0)



"""Custom JSON encoding support."""

from __future__ import absolute_import

import json
import sys


class EliotJSONEncoder(json.JSONEncoder):
    """JSON encoder with additional functionality.

    In particular, supports NumPy types.
    """

    def default(self, o):
        numpy = sys.modules.get("numpy", None)
        if numpy is not None:
            if isinstance(o, numpy.floating):
                return float(o)
            if isinstance(o, numpy.integer):
                return int(o)
            if isinstance(o, (numpy.bool, numpy.bool_)):
                return bool(o)
            if isinstance(o, numpy.ndarray):
                if o.size > 10000:
                    # Too big to want to log as-is, log a summary:
                    return {
                        "array_start": o.flat[:10000].tolist(),
                        "original_shape": o.shape,
                    }
                else:
                    return o.tolist()
        return json.JSONEncoder.default(self, o)


__all__ = ["EliotJSONEncoder"]



"""
Hypothesis strategies for eliot.
"""

from __future__ import unicode_literals

from functools import partial
from six import text_type as unicode

from hypothesis.strategies import (
    builds,
    dictionaries,
    fixed_dictionaries,
    floats,
    integers,
    lists,
    just,
    none,
    one_of,
    recursive,
    text,
    uuids,
)

from pyrsistent import pmap, pvector, ny, thaw

from .._action import (
    ACTION_STATUS_FIELD,
    ACTION_TYPE_FIELD,
    FAILED_STATUS,
    STARTED_STATUS,
    SUCCEEDED_STATUS,
    TaskLevel,
    WrittenAction,
)
from .._message import (
    EXCEPTION_FIELD,
    REASON_FIELD,
    TASK_LEVEL_FIELD,
    TASK_UUID_FIELD,
    WrittenMessage,
)

task_level_indexes = integers(min_value=1, max_value=10)
# Task levels can be arbitrarily deep, but in the wild rarely as much as 100.
# Five seems a sensible average.
task_level_lists = lists(task_level_indexes, min_size=1, max_size=6)
task_levels = task_level_lists.map(lambda level: TaskLevel(level=level))

# Text generation is slow, and most of the things are short labels. We set
# a restricted alphabet so they're easier to read, and in general large
# amount of randomness in label generation doesn't enhance our testing in
# any way, since we don't parse type names or user field values.
labels = text(min_size=1, max_size=8, alphabet="CGAT")

timestamps = floats(min_value=0, max_value=1000.0)

message_core_dicts = fixed_dictionaries(
    dict(
        task_level=task_level_lists.map(pvector),
        task_uuid=uuids().map(unicode),
        timestamp=timestamps,
    )
).map(pmap)

# Text generation is slow. We can make it faster by not generating so
# much. These are reasonable values.
message_data_dicts = dictionaries(
    keys=labels,
    values=labels,
    # People don't normally put much more than ten fields in their
    # messages, surely?
    max_size=10,
).map(pmap)


def written_from_pmap(d):
    """
    Convert a C{pmap} to a C{WrittenMessage}.
    """
    return WrittenMessage.from_dict(thaw(d))


def union(*dicts):
    result = pmap().evolver()
    for d in dicts:
        # Work around bug in pyrsistent where it sometimes loses updates if
        # they contain some kv pairs that are identical to the ones in the
        # dict being updated.
        #
        # https://github.com/tobgu/pyrsistent/pull/54
        for key, value in d.items():
            if key in result and result[key] is value:
                continue
            result[key] = value
    return result.persistent()


message_dicts = builds(union, message_data_dicts, message_core_dicts)
written_messages = message_dicts.map(written_from_pmap)

_start_action_fields = fixed_dictionaries(
    {ACTION_STATUS_FIELD: just(STARTED_STATUS), ACTION_TYPE_FIELD: labels}
)
start_action_message_dicts = builds(union, message_dicts, _start_action_fields).map(
    lambda x: x.update({TASK_LEVEL_FIELD: x[TASK_LEVEL_FIELD].set(-1, 1)})
)
start_action_messages = start_action_message_dicts.map(written_from_pmap)


def sibling_task_level(message, n):
    return message.task_level.parent().level.append(n)


_end_action_fields = one_of(
    just({ACTION_STATUS_FIELD: SUCCEEDED_STATUS}),
    fixed_dictionaries(
        {
            ACTION_STATUS_FIELD: just(FAILED_STATUS),
            # Text generation is slow. We can make it faster by not generating so
            # much. Thqese are reasonable values.
            EXCEPTION_FIELD: labels,
            REASON_FIELD: labels,
        }
    ),
)


def _make_written_action(start_message, child_messages, end_message_dict):
    """
    Helper for creating arbitrary L{WrittenAction}s.

    The child messages and end message (if provided) will be updated to have
    the same C{task_uuid} as C{start_message}. Likewise, their C{task_level}s
    will be such that they follow on from C{start_message}.

    @param WrittenMessage start_message: The message to start the action with.
    @param child_messages: A sequence of L{WrittenAction}s and
        L{WrittenMessage}s that make up the action.
    @param (PMap | None) end_message_dict: A dictionary that makes up an end
        message. If None, then the action is unfinished.

    @return: A L{WrittenAction}
    """
    task_uuid = start_message.task_uuid
    children = []

    for i, child in enumerate(child_messages, 2):
        task_level = TaskLevel(level=sibling_task_level(start_message, i))
        children.append(reparent_action(task_uuid, task_level, child))

    if end_message_dict:
        end_message = written_from_pmap(
            union(
                end_message_dict,
                {
                    ACTION_TYPE_FIELD: start_message.contents[ACTION_TYPE_FIELD],
                    TASK_UUID_FIELD: task_uuid,
                    TASK_LEVEL_FIELD: sibling_task_level(
                        start_message, 2 + len(children)
                    ),
                },
            )
        )
    else:
        end_message = None

    return WrittenAction.from_messages(start_message, children, end_message)


written_actions = recursive(
    written_messages,
    lambda children: builds(
        _make_written_action,
        start_message=start_action_messages,
        child_messages=lists(children, max_size=5),
        end_message_dict=builds(union, message_dicts, _end_action_fields) | none(),
    ),
)


def _map_messages(f, written_action):
    """
    Map C{f} across all of the messages that make up C{written_action}.

    This is a structure-preserving map operation. C{f} will be applied to all
    messages that make up C{written_action}: the start message, end message,
    and children. If any of the children are themselves L{WrittenAction}s, we
    recurse down into them.

    @param f: A function that takes a L{WrittenMessage} and returns a new
        L{WrittenMessage}.
    @param (WrittenAction | WrittenMessage) written_action: A written

    @return: A L{WrittenMessage} if C{written_action} is a C{WrittenMessage},
        a L{WrittenAction} otherwise.
    """
    if isinstance(written_action, WrittenMessage):
        return f(written_action)

    start_message = f(written_action.start_message)
    children = written_action.children.transform([ny], partial(_map_messages, f))
    if written_action.end_message:
        end_message = f(written_action.end_message)
    else:
        end_message = None

    return WrittenAction.from_messages(
        start_message=start_message, children=pvector(children), end_message=end_message
    )


def reparent_action(task_uuid, task_level, written_action):
    """
    Return a version of C{written_action} that has the given C{task_uuid} and
    is rooted at the given C{task_level}.

    @param UUID task_uuid: The new task UUID.
    @param TaskLevel task_level: The new task level.
    @param (WrittenAction | WrittenMessage) written_action: The action or
        message to update.

    @return: A new version of C{written_action}.
    """
    new_prefix = list(task_level.level)
    old_prefix_len = len(written_action.task_level.level)

    def fix_message(message):
        return message.transform(
            ["_logged_dict", TASK_LEVEL_FIELD],
            lambda level: new_prefix + level[old_prefix_len:],
        ).transform(["_logged_dict", TASK_UUID_FIELD], task_uuid)

    return _map_messages(fix_message, written_action)
